/* -------------------------------------------------------------- */
/*  LAST EDIT: Thu Oct 31 22:58:37 1996 by Adabas D Database (gutamil.v12-berlin.de!adabas)  */
/*  $Id: AdabasFormat.c,v 1.38 1996/10/31 22:22:00 adabas Exp $ */
/* -------------------------------------------------------------- */
/* Copyright (c) 1996 Christian Krone. All rights reserved.       */
/* This program is free software; you can redistribute it and/or  */
/* modify it under the same terms as Tcl itself.                  */
/* See also licence.terms                                         */
/* -------------------------------------------------------------- */

/* ----------------- Import interface --------------------------- */
#include <stdlib.h>
#include <string.h>
#include <tcl.h>

#include "adabas.h"
#include "sqlrte.h"

#include "AdabasUtil.h"
#include "AdabasPacket.h"

/* ----------------- Export interface --------------------------- */
#include "AdabasFormat.h"

/* ----------------- Type definitions --------------------------- */
#define DATA_TYPE_REGEXP_STRING "^([^(]*)\\(([0-9]*)(,(-?[0-9]*))?\\)$"
#define DATA_TYPE_RANGE 1
#define LENGTH_RANGE    2
#define FRAC_RANGE      4
#define VAL_POS        33
#define VAL_LEN        37

/* ----------------- Local procedure headings ------------------- */
static int checkGetParams _ANSI_ARGS_((Tcl_Interp *interp, int argc,
				       char **argv, data_type *dataType,
				       int *ioLen, int *len, int *frac,
				       int *startPos, char **nullValue));
static int putInt _ANSI_ARGS_((Tcl_Interp *interp, char *data, int ioLen,
			       int len, int startPos, char *nullValue));
static int getIoLen _ANSI_ARGS_((int dataType, int length));
static void longDescr2List _ANSI_ARGS_((Tcl_DString *substitute,
					long_descriptor *longDesc));
static int getEditForm _ANSI_ARGS_((Tcl_Interp *interp, char *value));
static int putEditForm _ANSI_ARGS_((Tcl_Interp *interp, char *value));
#if ADABAS_VERSION >= 62
static int appendUnicode _ANSI_ARGS_((Tcl_Interp *interp, char *data,
				      int length, int isElement));
#endif

/* ----------------- Exported procedures ------------------------- */

int AdabasCryptCmd (dummy, interp, argc, argv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
  Tcl_DString result;

  if (argc < 2) {
    interp->result = "wrong # args: should be \"adabas crypt clear_pw\"";
    return TCL_ERROR;      
  }
  
  Tcl_DStringInit   (&result);
  AdabasPasswdCrypt (&result, argv[1]);
  Tcl_DStringResult (interp, &result);
  return TCL_OK;
}

/* -------------------------------------------------------------- */

void AdabasPasswdCrypt (cryptPw, passwd)
     Tcl_DString *cryptPw;		/* resulting crypted password. */
     char        *passwd;		/* clear password to be cryptified. */
{
  name    clearname;
  cryptpw crypt_pw;

  c2p (clearname, passwd, sizeof (clearname));
  s02applencrypt (clearname, crypt_pw);
  decodeSequence (encEscape, cryptPw, crypt_pw, sizeof (crypt_pw));
}

/* -------------------------------------------------------------- */

int AdabasPutCmd (dummy, interp, argc, argv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int argc;				/* Number of arguments. */
    char **argv;			/* Argument strings. */
{
  data_type dataType  = DCHA;
  int       ioLen     = 2;
  int       len       = 1;
  int       startPos  = 1;
  int       frac      = 0;
  char     *nullValue = "";

  if (checkGetParams (interp, argc, argv, &dataType,
		      &ioLen, &len, &frac, &startPos, &nullValue) != TCL_OK)
    return TCL_ERROR;

  switch (dataType) {
#if ADABAS_VERSION >= 62
  case DFIXED:
  case DSMALLINT:
  case DINTEGER:
#else /* ADABAS_VERSION <= 61 */
  case CSP_INFO_FIXED:
  case CSP_INFO_SMALLINT:
  case CSP_INFO_INTEGER:
#endif
    return putInt (interp, argv[1], ioLen, len, startPos, nullValue);

  /*
   * Datatype UNKNOWN is a special flag, that the data should be
   * regarded as edit form.
   */
  case DUNKNOWN:
    Tcl_ResetResult (interp);
    return putEditForm (interp, argv[1]);

  default:
    Tcl_AppendResult (interp, " <", dataType2String (dataType),
		      ">", (char *) NULL);
    return TCL_OK;
  }
}

/* -------------------------------------------------------------- */

int AdabasGetCmd (dummy, interp, argc, argv)
     ClientData dummy;			/* Not used. */
     Tcl_Interp *interp;		/* Current interpreter. */
     int argc;				/* Number of arguments. */
     char **argv;			/* Argument strings. */
{
  data_type   dataType  = DCHA;
  int         ioLen     = 2;
  int         len       = 1;
  int         startPos  = 1;
  int         frac      = 0;
  char       *nullValue = "";
  Tcl_DString substitute;
  int         ret;

  if (checkGetParams (interp, argc, argv, &dataType,
		      &ioLen, &len, &frac, &startPos, &nullValue) != TCL_OK) {
    return TCL_ERROR;
  }

  /*
   * Datatype UNKNOWN is a special flag, that the data should be
   * regarded as query form.
   */
  if (dataType == DUNKNOWN)
    return getEditForm (interp, argv[1]);

  /*
   * Substitute the escaped characters of the given data and put a fixed
   * width string (with width 'ioLen') into 'substitute'.
   */
  Tcl_DStringInit (&substitute);
  ioLen = encodeSequence (interp, encEscape, "data", &substitute,
			  argv[1], startPos, ioLen);
  ret = AdabasAppendData (interp, Tcl_DStringValue (&substitute),
			  dataType, ioLen, len, frac, nullValue, 0);

  /*
   * For clean up free the resources, then return the result.
   */
  Tcl_DStringFree (&substitute);
  return ret;
}

/* -------------------------------------------------------------- */

int AdabasAppendData (interp, data, dataType, ioLen, len, frac,
		      nullValue, isElement)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *data;			/* Value to append as string. */
     int         dataType;              /* Data type of value. */
     int         ioLen;			/* IO-Length of value. */
     int         len;			/* Extern length of value. */
     int         frac;			/* Fraction, if value is a number. */
     char       *nullValue;		/* Representation of a NULL value. */
     int         isElement;		/* append result as element? */
{
  Tcl_DString     descString;
  long_descriptor longDesc;
  int4            intNumber;
  double          floatNumber;
  num_error       numError;
  char           *p;
  char            remC;
  char            numBuf[40];
  
  /*
   * Check, if the value to put is the NULL value.
   * If so, print the given nullValue (or nothing as default).
   */
  if (*(unsigned char *) data == 0xff) {
    if (isElement)
      Tcl_AppendElement (interp, nullValue ? nullValue : "");
    else
      Tcl_AppendResult  (interp, nullValue ? nullValue : "", (char *) NULL);
  } else {
    /*
     * Okay, it is a NOT NULL value to put. So now depending on the data type
     * we convert the value to string and return.
     */
    switch (dataType) {
      /*
       * Convert a vdn number without fraction into the 
       * string representation of an integer.
       */
#if ADABAS_VERSION >= 62
    case DFIXED:
    case DSMALLINT:
    case DINTEGER:
#else /* ADABAS_VERSION <= 61 */
    case CSP_INFO_FIXED:
    case CSP_INFO_SMALLINT:
    case CSP_INFO_INTEGER:
#endif
      if (!frac) {
	s40glint (data, 2, len, &intNumber, &numError);
	if (numError)
	  sprintf (numBuf, "<error %d>", numError);
	else
	  sprintf (numBuf, "%d", intNumber);
	if (isElement)
	  Tcl_AppendElement (interp, numBuf);
	else
	  Tcl_AppendResult  (interp, numBuf, (char *) NULL);
	break;
      } /* else nobreak; */
      
      /*
       * Convert a vdn number into the string representation of a float.
       */
#if ADABAS_VERSION >= 62
    case DFLOAT:
    case DVFLOAT:
#else /* ADABAS_VERSION <= 61 */
    case CSP_INFO_FLOAT:
    case CSP_INFO_EXPRESSION:
#endif
      s40glrel (data, 2, ioLen-1, &floatNumber, &numError);
      if (numError)
	sprintf (numBuf, "<error %d>", numError);
      else
	sprintf (numBuf, "%.*f", frac, floatNumber); 
      if (isElement)
	Tcl_AppendElement (interp, numBuf);
      else
	Tcl_AppendResult  (interp, numBuf, (char *) NULL);
      break;
    
      /*
       * Convert a fixed width char array into a right side trimmed string.
       */
#if ADABAS_VERSION >= 62
    case DCHA:
    case DCHE:
    case DCHB:
    case DDATE:
    case DTIME:
    case DTIMESTAMP:
    case DUNICODE:
    case DVARCHARA:
    case DVARCHARE:
    case DVARCHARB:
#else /* ADABAS_VERSION <= 61 */
    case CSP_INFO_CHAR:
    case CSP_INFO_VARCHAR:
    case CSP_INFO_DATE:
    case CSP_INFO_TIME:
    case CSP_INFO_TIMESTAMP:
#endif

#if ADABAS_VERSION >= 62
      if (*data == 1) {
	if (appendUnicode (interp, data+1, ioLen-1, isElement) == TCL_ERROR)
	  return TCL_ERROR;
	break;
      }
#endif

      /*
       * Find the last char in the string, which isn't a blank.
       * Then put the terminating null behind it, but remember the
       * old content of this cell, since it can be the start of the
       * next data...
       */
      for (p = data + ioLen; p-1 > data && *(p-1) == ' '; p--);
      remC = *p;
      *p   = 0;
      if (isElement)
	Tcl_AppendElement (interp, data+1);
      else
	Tcl_AppendResult  (interp, data+1, (char *) NULL);
      *p   = remC;
      break;
    
      /*
       * Convert a boolean (0/1) into its string representation.
       */
#if ADABAS_VERSION >= 62
    case DBOOLEAN:  
#else /* ADABAS_VERSION <= 61 */
    case CSP_INFO_BOOLEAN:
#endif
      if (isElement)
	Tcl_AppendElement (interp, *(data+1) ? "1" : "0");
      else
	Tcl_AppendResult  (interp, *(data+1) ? "1" : "0", (char *) NULL);
      break; 

#if ADABAS_VERSION >= 62
    case DSTRA:
    case DSTRB:
    case DSTRE:
    case DSTRUNI:
#else /* ADABAS_VERSION <= 61 */
    case CSP_INFO_C_OLDLONG_CHAR:
    case CSP_INFO_B_OLDLONG_BYTE:
    case CSP_INFO_A_OLDLONG_ASCII_DBYTE:
#endif
      /*
       * The data is a long descriptor.
       */
      memcpy (&longDesc, data+1, sizeof (longDesc));
      longDescr2List (&descString, &longDesc);
      if (isElement)
	Tcl_AppendElement (interp, Tcl_DStringValue (&descString));
      else
	Tcl_AppendResult  (interp, Tcl_DStringValue (&descString),
			   (char *) NULL);
      Tcl_DStringFree (&descString);
      break;

    default:
      Tcl_AppendResult (interp, " <", dataType2String (dataType),
			">", (char *) NULL);
      return TCL_ERROR;
    }
  }
  return TCL_OK;
}

/* ----------------- Local procedures --------------------------- */

static int checkGetParams (interp, argc, argv, dataType, ioLen, len, frac,
			   startPos, nullValue)
     Tcl_Interp *interp;		/* Current interpreter. */
     int         argc;			/* Number of arguments. */
     char      **argv;			/* Argument strings. */
     data_type  *dataType;		/* Data type of value. */
     int        *ioLen;			/* IO-Length of value. */
     int        *len;			/* Extern length of value. */
     int        *frac;			/* Fraction, if value is a number. */
     int        *startPos;		/* Startposition in data part. */
     char      **nullValue;		/* Representation of a NULL value. */
{
  size_t      length;
  int         c;
  Tcl_DString dStr;
  char       *start, *end;
  Tcl_RegExp  dataTypeRegexp;

  if (argc < 2) {
    Tcl_AppendResult (interp, "wrong # args: should be \"adabas ", argv[0],
		      " string ?data_type? ?option ...?\"", (char *) NULL);
    return TCL_ERROR;
  }
  if (argc >= 3 && argv[2][0] != '-') {
    /*
     * First compile the regular expression for 'datatype(len,frac?)'.
     * Since tcl maintains a small cache of compiled regexp, there are
     * good chances, that no new compilation is neccassary...
     */
    dataTypeRegexp = Tcl_RegExpCompile (interp, DATA_TYPE_REGEXP_STRING);
    if (dataTypeRegexp == (Tcl_RegExp) NULL)
      return TCL_ERROR;

    switch (Tcl_RegExpExec (interp, dataTypeRegexp, argv[2], argv[2])) {
    case -1:
      return TCL_ERROR;

    case 1:
      Tcl_DStringInit (&dStr);

      Tcl_RegExpRange (dataTypeRegexp, DATA_TYPE_RANGE, &start, &end);
      Tcl_DStringAppend (&dStr, start, end-start);
      *dataType = scanDataType (interp, Tcl_DStringValue (&dStr));
      Tcl_DStringFree (&dStr);

      Tcl_RegExpRange (dataTypeRegexp, LENGTH_RANGE, &start, &end);
      Tcl_DStringAppend (&dStr, start, end-start);
      if (Tcl_GetInt (interp, Tcl_DStringValue (&dStr), len) == TCL_ERROR)
	return TCL_ERROR;
      Tcl_DStringFree (&dStr);

      Tcl_RegExpRange (dataTypeRegexp, FRAC_RANGE, &start, &end);
      if (start) {
	Tcl_DStringAppend (&dStr, start, end-start);
	if (Tcl_GetInt (interp, Tcl_DStringValue (&dStr), frac) == TCL_ERROR)
	  return TCL_ERROR;
	Tcl_DStringFree (&dStr);
      }
      break;

    case 0:
      *dataType = scanDataType (interp, argv[2]);
      if (*dataType == DUNKNOWN) {
	if (!strncmp (argv[2], "editform", strlen (argv[2])))
	  return TCL_OK;
	return TCL_ERROR;
      }
      break;
    }

    argc--; argv++;
  }
  
  for (argv+=2, argc-=2; argc > 1; argv += 2, argc -= 2) {
    c      = argv[0][1];
    length = strlen (argv[0]);
    
    if (c == 'i' && !strncmp (argv[0], "-ioLen", length) && length>3) {
      if (Tcl_GetInt (interp, argv[1], ioLen)    == TCL_ERROR)
	return TCL_ERROR;
    } else if (c == 'l' && !strncmp (argv[0], "-len", length)) {
      if (Tcl_GetInt (interp, argv[1], len)      == TCL_ERROR)
	return TCL_ERROR;
    } else if (c == 'f' && !strncmp (argv[0], "-frac", length)) {
      if (Tcl_GetInt (interp, argv[1], frac)     == TCL_ERROR)
	return TCL_ERROR;
    } else if (c == 'b' && !strncmp (argv[0], "-bufPos", length)) {
      if (Tcl_GetInt (interp, argv[1], startPos) == TCL_ERROR)
	return TCL_ERROR;
    } else if (c == 'm' && !strncmp (argv[0], "-mode", length)) {
    } else if (c == 'n' && !strncmp (argv[0], "-nullValue", length)) {
      *nullValue = argv[1];
    } else if (c == 'i' && !strncmp (argv[0], "-ioType", length) && length>3) {
    } else {
      Tcl_AppendResult (interp, "bad option \"", argv[0], "\": should be ", 
			"-ioLen, -len, -frac, -bufPos, -ioType or -mode",
			(char *) NULL);
      return TCL_ERROR;
    }
  }
  if (argc == 1) {
    Tcl_AppendResult (interp, "value wanted for option \"", argv[0],
		     "\"", (char *) NULL);
    return TCL_ERROR;
  }

  /*
   * Some datatypes have default ioLens. So if they are not given
   * explicitly, here we can assign them.
   */
  if (*ioLen < 0)
    *ioLen = getIoLen (*dataType, *len);

  return TCL_OK;
}

/* -------------------------------------------------------------- */

static int putInt (interp, data, ioLen, len, startPos, nullValue)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *data;			/* Value to append as string. */
     int         ioLen;			/* IO-Length of value. */
     int         len;			/* Extern length of value. */
     int         startPos;		/* Startposition in data part. */
     char       *nullValue;		/* Representation of a NULL value. */
{
  Tcl_DString result;
  num_error   num_error;
  int         intNumber;
  char        vdnNumber[12];

  if (Tcl_GetInt (interp, data, &intNumber) == TCL_ERROR)
    return TCL_ERROR;

  s41plint (vdnNumber, 1, len, 0, intNumber, &num_error);
  if (num_error)
    sprintf (interp->result, "putInt conversion error %d", num_error);
  else {
    Tcl_DStringInit (&result);
    decodeSequence (encEscape, &result, "", 1); /* defined byte */
    decodeSequence (encEscape, &result, vdnNumber, ioLen-1);
    Tcl_DStringResult (interp, &result);
  }
  return TCL_OK;
}

/* -------------------------------------------------------------- */

static int getIoLen (dataType, length)
     int dataType;			/* Data type to get IO length of. */
     int length;			/* Extern length of value. */
{
  switch (dataType) {

  case DSMALLINT:
    return 5;

  case DFIXED:
  case DINTEGER:
  case DFLOAT:
  case DVFLOAT:
    return 11;

  case DSTRA:
  case DSTRE:
  case DSTRUNI:
    return 41;

  case DBOOLEAN:
    return 2;

  case DDATE:
    return 11;

  case DTIME:
    return 11;

  case DTIMESTAMP:
    return 21;

  case DUNICODE:
    return 2*length+1;

  case DCHA:
  case DCHB:
  case DCHE:
  case DVARCHARA:
  case DVARCHARB:
  case DVARCHARE:
  default:
    return length+1;
  }
}

/* -------------------------------------------------------------- */

static void longDescr2List (substitute, longDesc)
     Tcl_DString     *substitute;	/* Resulting Char representation. */
     long_descriptor *longDesc;		/* Long descriptor to convert. */
{
  char        number[13];
  char       *valmode;
  Tcl_DString hexStr;

  Tcl_DStringInit (substitute);
  Tcl_DStringInit (&hexStr);

  decodeSequence (encHex, &hexStr, longDesc->descriptor,
		  sizeof (longDesc->descriptor));
  Tcl_DStringAppendElement (substitute, Tcl_DStringValue (&hexStr));
  Tcl_DStringFree (&hexStr);

  decodeSequence (encHex, &hexStr, longDesc->tabid,
		  sizeof (longDesc->tabid));
  Tcl_DStringAppendElement (substitute, Tcl_DStringValue (&hexStr));
  Tcl_DStringFree (&hexStr);

  decodeSequence (encHex, &hexStr, longDesc->serverdb_no,
		  sizeof (longDesc->serverdb_no));
  Tcl_DStringAppendElement (substitute, Tcl_DStringValue (&hexStr));
  Tcl_DStringFree (&hexStr);

  Tcl_DStringAppendElement (substitute,
			    longDesc->replicated ? "true" : "false");

  sprintf (number, "%d", longDesc->change);
  Tcl_DStringAppendElement (substitute, number);

  sprintf (number, "%d", longDesc->intern_pos);
  Tcl_DStringAppendElement (substitute, number);

  sprintf (number, "%d", longDesc->infoset);
  Tcl_DStringAppendElement (substitute, number);

  Tcl_DStringAppendElement (substitute,
			    longDesc->short_scol ? "true" : "false");

  switch (longDesc->valmode) {
  case VM_DATAPART:     valmode = "datapart";   break;
  case VM_ALLDATA:      valmode = "alldata";    break;
  case VM_LASTDATA:     valmode = "lastdata";   break;
  case VM_NODATA:       valmode = "nodata";     break;
  case VM_NO_MORE_DATA: valmode = "noMoreData"; break;
  case VM_LAST_PUTVAL:  valmode = "lastPutval"; break;
  case VM_DATA_TRUNC:   valmode = "dataTrunc";  break;
  case VM_CLOSE:        valmode = "close";      break;
  default:              valmode = "???";        break;
  }
  Tcl_DStringAppendElement (substitute, valmode);

  sprintf (number, "%d", longDesc->valind);
  Tcl_DStringAppendElement (substitute, number);

  sprintf (number, "%d", longDesc->valpos);
  Tcl_DStringAppendElement (substitute, number);

  sprintf (number, "%d", longDesc->vallen);
  Tcl_DStringAppendElement (substitute, number);
}

/* -------------------------------------------------------------- */

int list2LongDescr (interp, inDescrString, outLongDesc)
     Tcl_Interp      *interp;		/* Current interpreter. */
     char            *inDescrString;	/* char representation of long desc. */
     long_descriptor *outLongDesc;	/* resulting long descriptor. */
{
  int     argc;
  char  **argv;
  int     number;
  int     length;
  valmode valmode = VM_NO_MORE_DATA;
  Tcl_DString tmpStr;

  if (Tcl_SplitList (interp, inDescrString, &argc, &argv) == TCL_ERROR)
     return TCL_ERROR;
  if (argc != 12) {
     ckfree ((char *) argv);
     interp->result = "No valid long descriptor given";
     return TCL_ERROR;
  }

  Tcl_DStringInit (&tmpStr);
  if (encodeSequence (interp, encHex, "descriptor", &tmpStr, argv[0], 1,
		      sizeof (outLongDesc->descriptor)) == TCL_ERROR)
    goto err;
  memcpy (outLongDesc->descriptor, Tcl_DStringValue (&tmpStr),
	  sizeof (outLongDesc->descriptor));
  Tcl_DStringFree (&tmpStr);

  if (encodeSequence (interp, encHex, "tabid", &tmpStr, argv[1], 1,
		      sizeof (outLongDesc->tabid)) == TCL_ERROR)
    goto err;
  memcpy (outLongDesc->tabid, Tcl_DStringValue (&tmpStr),
	  sizeof (outLongDesc->tabid));
  Tcl_DStringFree (&tmpStr);

  if (encodeSequence (interp, encHex, "serverdbNo", &tmpStr, argv[2], 1,
		      sizeof (outLongDesc->serverdb_no)) == TCL_ERROR)
    goto err;
  memcpy (outLongDesc->serverdb_no, Tcl_DStringValue (&tmpStr),
	  sizeof (outLongDesc->serverdb_no));
  Tcl_DStringFree (&tmpStr);

  if (Tcl_GetBoolean (interp, argv[3], &number) == TCL_ERROR)
    goto err;
  outLongDesc->replicated = number;

  if (Tcl_GetInt (interp, argv[4], &number) == TCL_ERROR)
    goto err;
  outLongDesc->change = number;

  if (Tcl_GetInt (interp, argv[5], &number) == TCL_ERROR)
    goto err;
  outLongDesc->intern_pos = number;

  if (Tcl_GetInt (interp, argv[6], &number) == TCL_ERROR)
    goto err;
  outLongDesc->infoset = number;

  if (Tcl_GetBoolean (interp, argv[7], &number) == TCL_ERROR)
    goto err;
  outLongDesc->short_scol = number;

  length = strlen (argv[8]);
  switch (argv[8][0]) {
  case 'a':
    if (!strncmp (argv[8], "alldata", length) && length > 0)
      valmode = VM_ALLDATA;
      break;
  case 'c':
    if (!strncmp (argv[8], "close", length) && length > 0)
      valmode = VM_CLOSE;
      break;
  case 'd':
    if (!strncmp (argv[8], "datapart", length) && length>0)
      valmode = VM_DATAPART;
    else if (!strncmp (argv[8], "dataTrunc", length) && length>0)
      valmode = VM_DATA_TRUNC;
    break;
  case 'l':
    if (!strncmp (argv[8], "lastdata", length) && length>0)
      valmode = VM_LASTDATA;
    else if (!strncmp (argv[8], "lastPutval", length) && length>0)
      valmode = VM_LAST_PUTVAL;
    break;
  case 'n':
    if (!strncmp (argv[8], "nodata", length) && length>0)
      valmode = VM_NODATA;
    else if (!strncmp (argv[8], "noMoreData", length) && length>0)
      valmode = VM_NO_MORE_DATA;
    break;
  default:
    Tcl_AppendResult (interp, "Expected valmode, but got \"",
		      argv[8], "\"", (char *) NULL);
    goto err;
  }
  outLongDesc->valmode    = valmode;

  if (Tcl_GetInt (interp, argv[9], &number) == TCL_ERROR)
    goto err;
  outLongDesc->valind       = number;

  if (Tcl_GetInt (interp, argv[10], &number) == TCL_ERROR)
    goto err;
  outLongDesc->valpos    = number;

  if (Tcl_GetInt (interp, argv[11], &number) == TCL_ERROR)
    goto err;
  outLongDesc->vallen    = number;

  ckfree ((char *) argv);
  return TCL_OK;

 err:
  ckfree ((char *) argv);
  return TCL_ERROR;
}

/* -------------------------------------------------------------- */
#if ADABAS_VERSION >= 62

static int appendUnicode (interp, data, length, isElement)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *data;			/* value to append (w/o defbyte). */
     int         length;		/* length of value. */
     int         isElement;		/* append result as element? */
{
  Tcl_DString uniBuf;
  int4        destLen = length;

  Tcl_DStringInit (&uniBuf);
  Tcl_DStringSetLength (&uniBuf, length);

  if (AdabasFromUnicode (interp, data, length, Tcl_DStringValue (&uniBuf),
			 destLen) == TCL_ERROR)
    return TCL_ERROR;

  Tcl_DStringSetLength (&uniBuf, destLen);

  if (isElement)
    Tcl_AppendElement (interp, Tcl_DStringValue (&uniBuf));
  else
    Tcl_AppendResult  (interp, Tcl_DStringValue (&uniBuf), (char *) NULL);

  Tcl_DStringFree (&uniBuf);
  return TCL_OK;
}
#endif /* ADABAS_VERSION >= 62 */

/* -------------------------------------------------------------- */

/*
 * The procedures getEditForm and putEditForm are for transforming
 * a valid tcl list (containing e.g. some lines from a text widget)
 * into a so called editform, the internal format for storing queries
 * into a LONG BYTE column, and back.
 * The intern format is:
 * - one byte: the length of the line (without surrounding blanks);
 * - one byte: the number of leading blanks;
 * - the trimmed line.
 */

static int getEditForm (interp, value)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *value;			/* internal format of edit form. */
{
  int         ix;
  int         ioLen;
  int         lineLength;
  int         cntBlanks;
  Tcl_DString substitute;
  Tcl_DString editForm;
  char        blankBuf[256];
  char       *str;

  Tcl_DStringInit (&editForm);
  
  ioLen = encodeSequence (interp, encEscape, "editForm", &substitute,
			  value, 1, strlen (value));

  for (ix = 0; ix < ioLen; ix += lineLength) {
    str        = Tcl_DStringValue (&substitute) + ix;
    lineLength = * (unsigned char *) str++;
    cntBlanks  = * (unsigned char *) str++;

    if (ix + lineLength > ioLen) {
      sprintf (interp->result, "Invalid query form at position %d", ix);
      Tcl_DStringFree (&substitute);
      Tcl_DStringFree (&editForm);
      return TCL_ERROR;
    }

    Tcl_DStringStartSublist (&editForm);
    if (cntBlanks) {
      memset (blankBuf, ' ', cntBlanks);
      Tcl_DStringAppend (&editForm, blankBuf, cntBlanks);
    }
    Tcl_DStringAppend (&editForm, str, lineLength - 2);
    Tcl_DStringEndSublist (&editForm);
  }

  Tcl_DStringFree (&substitute);
  Tcl_DStringResult (interp, &editForm);

  return TCL_OK;
}

/* -------------------------------------------------------------- */

static int putEditForm (interp, value)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *value;			/* External format of query. */
{
  int           lineLength;
  int           cntBlanks;
  Tcl_DString   editForm;
  int           argc;
  char        **argv;
  char        **curr;
  char         *str;
  unsigned char info[2];

  Tcl_DStringInit (&editForm);

  if (Tcl_SplitList (interp, value, &argc, &argv) == TCL_ERROR)
    return TCL_ERROR;

  for (curr = argv; *curr; curr++) {
    for (str = *curr, cntBlanks = 0; *str == ' '; str++, cntBlanks++);
    for (lineLength = strlen (str); str[lineLength-1] == ' '; lineLength--);
    if (cntBlanks > 255 || lineLength > 255) {
      interp->result = "Too long line in query form";
      ckfree ((char *) argv);
      Tcl_DStringFree (&editForm);
      return TCL_ERROR;
    }
    info[0] = (unsigned char) lineLength + 2;
    info[1] = (unsigned char) cntBlanks;
    decodeSequence (encEscape, &editForm, (char *) info, sizeof (info));
    Tcl_DStringAppend (&editForm, str, lineLength);
  }

  Tcl_DStringResult (interp, &editForm);
  ckfree ((char *) argv);
  return TCL_OK;
}

/* ---------------------------------------------------------------------
 * Some stuff for editing with emacs...
 * ---------------------------------------------------------------------
 * Local Variables: 
 * mode: c
 * mode: font-lock
 * End: 
 */
