/* -------------------------------------------------------------- */
/*  LAST EDIT: Thu Oct 10 22:01:58 1996 by Adabas D Database (gutamil.v12-berlin.de!adabas)  */
/*  $Id: AdabasSend.h,v 1.18 1996/10/13 13:03:06 adabas Exp $ */
/* -------------------------------------------------------------- */
/* Copyright (c) 1996 Christian Krone. All rights reserved.       */
/* This program is free software; you can redistribute it and/or  */
/* modify it under the same terms as Tcl itself.                  */
/* See also licence.terms                                         */
/* -------------------------------------------------------------- */

/* ----------------- Constant definitions ----------------------- */
#define ADABAS_TCL_VERSION  0.5

/*
 * The name of the global array, where the results of the adasql
 * commands get stored, and its indexes.
 */
#define ADABAS_MESSAGES     "adamsg"
#define ADAMSG_VERSION      "version"
#define ADAMSG_NULLVALUE    "nullvalue"
#define ADAMSG_HANDLE       "handle"
#define ADAMSG_ERRORPOS     "errorpos"
#define ADAMSG_ERRORTXT     "errortxt"
#define ADAMSG_ROWS         "rows"
#define ADAMSG_INTO_VARS    "intoVars"
#define ADAMSG_RC           "rc"
#define ADAMSG_COLTYPES     "coltypes"
#define ADAMSG_COLLENGTHS   "collengths"
#define ADAMSG_COLPRECS     "colprecs"
#define ADAMSG_COLSCALES    "colscales"
#define ADAMSG_RC           "rc"
#define ADAMSG_RC           "rc"
#define ADAMSG_RC           "rc"

/*
 * Access to the array "adamsg" should go only over this macros.
 */
#define AdabasSetMsg(interp,index,value) \
    Tcl_SetVar2 (interp, ADABAS_MESSAGES, index, value, TCL_GLOBAL_ONLY)
#define AdabasUnsetMsg(interp,index) \
    Tcl_UnsetVar2 (interp, ADABAS_MESSAGES, index, TCL_GLOBAL_ONLY)
#define AdabasGetMsg(interp,index) \
    Tcl_GetVar2 (interp, ADABAS_MESSAGES, index, TCL_GLOBAL_ONLY)
#define AdabasLAppendMsg(interp,index,value) \
    Tcl_SetVar2 (interp, ADABAS_MESSAGES, index, value, \
		 TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT)

/* ----------------- Type definitions --------------------------- */

  /*
   * In pre-6.2 we have no part kinds, so here we have to define our
   * own enumeration type. Of cource we only define the kinds, we
   * actually use. And since we try to make life easy, we define
   * this enum for 6.2, too...
   */
#define PART_KINDS 6
typedef enum {
  pkNil,
  pkCommand,
  pkData,
  pkResultcount,
  pkLongdata,
  pkParsid
} partKind;

typedef struct {
  char *buf;
  int4  bufLength;
  int   subst;
} AdabasPartInfo;

/*
 * Note, that it is important to put the array of records at the top,
 * since else the compiler on an HP-UX generated code, that accesses
 * the wrong entries.
 */
typedef struct {
  AdabasPartInfo   partInfos[PART_KINDS];
  AdabasRteInfo    rteInfo;
  AdabasPacketInfo packetInfo;

#if ADABAS_VERSION >= 62
  /*
   * In 6.2 the packet is allocated by the rte, so all we have to do here
   * is remember the address of it. The address of the return packet we
   * store here too, even if it doesn't change for 6.1...
   */
  packet *sndPacket;
#else /* ADABAS_VERSION <= 61 */
  /*
   * In pre-6.2 we have to declare the space for the packet ourself.
   * This is exactly the right (and the only one) place for this...
   */
  packet sqlPacket;
#endif
  /*
   * The address of the return packet we store here too, even if it
   * doesn't change for 6.1...
   */
  packet *rcvPacket;
} AdabasInfo;

/* ----------------- Makro definitions -------------------------- */
#define SET_VERSION(result)					\
sprintf (result, "AdabasTcl %.1f (for Adabas D %01d.%01d)",	\
	 ADABAS_TCL_VERSION, ADABAS_VERSION/10, ADABAS_VERSION%10)

/* ----------------- Exported procedure headings ---------------- */
extern int AdabasOpenConnection _ANSI_ARGS_((Tcl_Interp *interp,
					     int service, char *serverdb,
					     char *name, AdabasInfo **vars));
extern int AdabasCloseConnection _ANSI_ARGS_((Tcl_Interp *interp, char *Name,
					      AdabasInfo *vars));
extern int AdabasLogonHandle _ANSI_ARGS_((Tcl_Interp *interp, char *logonName,
					  AdabasInfo **logon));
extern int AdabasSendCmd _ANSI_ARGS_((ClientData clientData,
				      Tcl_Interp *interp,
				      int argc, char **argv));
extern int AdabasFillPacket _ANSI_ARGS_((Tcl_Interp *interp,
					 AdabasInfo *vars,
					 AdabasPacketInfo *thisOptions));
extern int AdabasSendPacket _ANSI_ARGS_((Tcl_Interp *interp,
					 AdabasInfo *vars));
extern int AdabasSend2PartPacket _ANSI_ARGS_((Tcl_Interp *interp,
					      AdabasInfo *vars,
					      int part1Kind, char *part1,
					      int part1Int, int part2Kind,
					      char *part2, int part2Int));
extern int AdabasSendCmdPacket _ANSI_ARGS_((Tcl_Interp *interp,
					    AdabasInfo *vars, int withInfo,
					    int massCnt, char *command));

/* ---------------------------------------------------------------------
 * Some stuff for editing with emacs...
 * ---------------------------------------------------------------------
 * Local Variables: 
 * mode: c
 * mode: font-lock
 * End: 
 */
