/* -------------------------------------------------------------- */
/*  LAST EDIT: Thu Oct 31 23:04:12 1996 by Adabas D Database (gutamil.v12-berlin.de!adabas)  */
/*  $Id: AdabasUtil.c,v 1.16 1996/10/31 22:21:33 adabas Exp $ */
/* -------------------------------------------------------------- */
/* Copyright (c) 1996 Christian Krone. All rights reserved.       */
/* This program is free software; you can redistribute it and/or  */
/* modify it under the same terms as Tcl itself.                  */
/* See also licence.terms                                         */
/* -------------------------------------------------------------- */

/* ----------------- Import interface --------------------------- */
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <tcl.h>

#include "adabas.h"
#include "sqlrte.h"

#define BASE64 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"

/* ----------------- Export interface --------------------------- */
#include "AdabasUtil.h"

/* ----------------- Local variables ---------------------------- */
#if ADABAS_VERSION >= 62
static encodings_ptr encodings;
static int2          encoding_cnt;
static int2          extCodeset;
static int2          ucsCodeset;
#endif

/* ----------------- Exported procedures ------------------------- */

void p2c (c_out, p_in, len)
     char *c_out;			/* Resulting null terminated string. */
     char *p_in;			/* Incoming fixed width string. */
     int   len;				/* Length of p_in. */
{
  char *pptr = p_in+len-1; /* pascal string ptr */
  char *cptr = c_out+len;  /* C      string ptr */
  
  *cptr -- = '\0';

  if (!len) return;
  
  while (pptr != p_in && * pptr == ' ') {
    *cptr-- = '\0';
    pptr--;
  }
  
  if (pptr == p_in && *pptr == ' ')
    *cptr-- = '\0';
  else {
    while (pptr != p_in)
      *cptr-- = *pptr--;
    *cptr = *pptr; /* don't forget last char */
  }
}

/* -------------------------------------------------------------- */

void c2p (p_out, c_in, len)
     char *p_out;			/* Resulting fixed width string. */
     char *c_in;			/* Incoming null terminated string. */
     int   len;				/* Length of p_out. */
{
  while (len-- > 0) {
    if (*c_in == '\0')
      *p_out++ = ' ';
    else
      *p_out++ = *c_in++;
  }
}

/* -------------------------------------------------------------- */

DataEncoding scanEncoding (interp, encoding)
     Tcl_Interp *interp;		/* Current interpreter. */
     char       *encoding;		/* String to scan for long encoding. */
{
  int c      = encoding[0];
  int length = strlen (encoding);

  switch (c) {
  case 'b':
    if (!strncmp (encoding, "base64", length))
      return encBase64;
  case 'e':
    if (!strncmp (encoding, "escape", length))
      return encEscape;
  case 'h':
    if (!strncmp (encoding, "hex", length))
      return encEscape;
  default:
    Tcl_AppendResult (interp, "invalid encoding \"", encoding, "\", must be ",
		      "base64, escape or hex", (char *) NULL);
    return encNil;
  }
}

/* -------------------------------------------------------------- */

void decodeSequence (encoding, out, in, length)
     int encoding;			/* Wanted encoding. */
     Tcl_DString *out;			/* Resulting decoded string. */
     char        *in;			/* Incoming data. */
     int          length;		/* Length of in. */
{
  char  x64[73];
  int   i;
  int   j;
  char *base64;
  unsigned char *x;
  unsigned char *p;
  char oneEscape[5];
  char oneHex[3];

  switch (encoding) {

  case encNil:
    Tcl_DStringAppend (out, in, length);
    return;

  case encBase64:
    j      = 0;
    base64 = BASE64;
    x      = (unsigned char *) in;

    for (i = 0; i < length; i+=3) {
      x64[j++] = base64[                     (x[i  ] >> 2)];
      x64[j++] = base64[((x[i  ]& 3) << 4) + (x[i+1] >> 4)];
      if (i == length-2) break;
      x64[j++] = base64[((x[i+1]&15) << 2) + (x[i+2] >> 6)];
      if (i == length-1) break;
      x64[j++] = base64[ (x[i+2]&63)                       ];

      if (j == sizeof (x64) - 1) {
	x64[j] = 0;
	Tcl_DStringAppendElement (out, x64);
	j = 0;
      }
    }
    if (j) {
      x64[j] = 0;
      Tcl_DStringAppendElement (out, x64);
    }
    return;

  case encEscape:
    x = (unsigned char *) in;
    p = (unsigned char *) in;

    while (length--) {
      if (isprint (*x) && *x != '\\')
	x++;
      else {
	if (p != x) Tcl_DStringAppend (out, (char *) p, x-p);
	sprintf (oneEscape, "\\%03o", *(x++));
	Tcl_DStringAppend (out, oneEscape, 4);
	p = x;
      }
    }
    if (p != x) Tcl_DStringAppend (out, (char *) p, x-p);
    return;

  case encHex:
    x = (unsigned char *) in;

    while (length-- > 0) {
      sprintf (oneHex, "%02x", *(x++));
      Tcl_DStringAppend (out, oneHex, 2);
    }
    return;

  default:
    return;
  }
}

/* -------------------------------------------------------------- */

int4 encodeSequence (interp, encoding, what, out, in, inPos, length)
     Tcl_Interp  *interp;		/* Current interpreter. */
     int          encoding;		/* Encoding, data is decoded. */
     char        *what;			/* specification (for errormsg). */
     Tcl_DString *out;			/* Resulting encoded string. */
     char        *in;			/* incoming decoded string. */
     int          inPos;		/* Start position (counts from 1). */
     int          length;		/* length of in. */
{
  char *p;
  char *old;
  int   count;
  int4  resLength;
  char  c;
  char *curr;
  int   numb;
  unsigned char u;

  switch (encoding) {

  case encNil:
    Tcl_DStringInit (out);
    Tcl_DStringAppend (out, in+inPos-1, length);
    return length;

  case encBase64:
    Tcl_DStringInit (out);
    Tcl_DStringAppend (out, "base64 not yet implemented", 26);
    return 26;

  case encEscape:
    p         = in;
    resLength = 0;

    /* First find the start position. This can't be done by pointer
     * arithmetic, since all the special symbols are escaped yet.
     */
    while (*p != 0 && inPos-- > 1) {
      if (*p == '\\') {
	c  = Tcl_Backslash (p, &count);
	p += count;
      } else
	p++;
    }

    Tcl_DStringInit (out);
    old = p;
    while (*p != 0 && (!length || resLength < length)) {
      if (*p == '\\') {
	if (p != old)
	  Tcl_DStringAppend (out, old, p-old);
	c   = Tcl_Backslash (p, &count);
	Tcl_DStringAppend (out, &c, 1);
	resLength++;
	p  += count;
	old = p;
      } else {
	p++; resLength++;
      }
    }
    if (p != old)
      Tcl_DStringAppend(out, old, p-old);    
    return resLength;

  case encHex:
    curr = in;

    while (length-- > 0) {
      if (sscanf (curr, "%2x", &numb) != 1) {
	Tcl_AppendResult (interp, "Expected ", what, ", but got \"",
			  in, "\"", (char *) NULL);
	return -1;
      }
      u = (unsigned char) numb;
      Tcl_DStringAppend (out, (char *) &u, 1);
      curr += 2;
    }
    return length/2;

  default:
    interp->result = "Invalid encoding ???";
    return -1;
  }
}

#if ADABAS_VERSION >= 62
int AdabasFromUnicode (interp, src, srcLen, dest, destLen)
     Tcl_Interp  *interp;		/* Current interpreter. */
     char        *src;
     int4         srcLen;
     char        *dest;
     int4         destLen;
{
  uni_error rc;
  int4      errCharNo;
  c40       msg;
  char      c_msg[41];

  if (!encodings && AdabasUtilInit (interp) == TCL_ERROR)
    return TCL_ERROR;

  s80uni_trans (encodings, encoding_cnt,
		src,  srcLen,   ucsCodeset,
		dest, &destLen, extCodeset,
		0, &rc, &errCharNo);

  if (rc != UNI_OK) {
    s80uni_error (rc, msg);
    p2c (c_msg, msg, sizeof (msg));
    Tcl_SetResult (interp, c_msg, TCL_VOLATILE);
    return TCL_ERROR;
  } else
    return TCL_OK;
}

static int AdabasUtilInit (interp)
     Tcl_Interp  *interp;		/* Current interpreter. */
{
  uni_load_error  loadRc;
  c40             msg;
  char            c_msg[41];
  char *iso8859_1 = "ISO8859_1         ";
  int2  swapIt    = 1;
  char *pointIt   = (char *) &swapIt;

  ucsCodeset = pointIt[0] ? UNICODE_SWAP : UNICODE;

  s82uni_fload (&encodings, &encoding_cnt, 1,
		iso8859_1, &extCodeset, &loadRc);

  if (loadRc == UNI_NO_DBLANG_FOUND)
    s82uni_fload (&encodings, &encoding_cnt, 0,
		  iso8859_1, &extCodeset, &loadRc);

  if (loadRc != UNI_LOAD_OK) {
    s82uni_load_error (loadRc, msg);
    p2c (c_msg, msg, sizeof (msg));
    Tcl_SetResult (interp, c_msg, TCL_VOLATILE);
    return TCL_ERROR;
  }

  return TCL_OK;
}
#endif /* ADABAS_VERSION >= 62 */

/* ---------------------------------------------------------------------
 * Some stuff for editing with emacs...
 * ---------------------------------------------------------------------
 * Local Variables: 
 * mode: c
 * mode: font-lock
 * End: 
 */
