# ---------------------------------------------------------------------
#  LAST EDIT: Thu Oct 24 20:59:07 1996 by Adabas D Database (gutamil.v12-berlin.de!adabas) 
#  $Id: panel.tcl,v 1.28 1996/10/24 19:51:59 adabas Exp $
# ---------------------------------------------------------------------

# ---------------------------------------------------------------------
# createPanel: Creates a panel canvas with a lot of objects on it.
#              Note the traffic light near the middle.
# ---------------------------------------------------------------------
proc createPanel {} {
    global adabastcl_library useNetscape

    set bitmaps $adabastcl_library/bitmaps

    option add *panel.edition      "Database Server" widgetDefault
    option add *panel.netscapeHelp False             widgetDefault

    image create photo panel -file $bitmaps/panel.gif
    canvas .panel -highlightthickness 0 \
	    -width [image width panel] -height [image height panel]
    .panel create image 0 0 -image panel -anchor nw  -tag bg
    set width  [image width  panel]
    set height [image height panel]

    .panel create text  [expr {$width-80}] 12 \
	    -text "Adabas D" -fill black  -anchor n -tag adabasD
    .panel create text  [expr {$width-82}] 10 \
	    -text "Adabas D" -fill red -anchor n -tag adabasD
    trySetFont .panel adabasD \
	    -Adobe-Times-Bold-R-Normal--34-*-*-*-*-*-*-* \
	    -Adobe-Times-Bold-R-Normal--24-*-*-*-*-*-*-*

    .panel create text  [expr {$width-80}] 50 \
	    -text [option get .panel edition Edition] -fill red -tag edition
    trySetFont .panel edition -Adobe-Times-Bold-I-Normal--14-*-*-*-*-*-*-*

    pack .panel
    wm title     . "Adabas D"
    wm resizable . 0 0
    update

    .panel create text 10 60                  -fill orange -anchor w -tag msg
    .panel create text  7 [expr {$height-10}] -fill blue   -anchor w -tag dbState

    foreach labSpec {
	{Info blue  120}
	{Help green  85}
	{Quit red    50}
    } {
	set label [lindex $labSpec 0]
	set color [lindex $labSpec 1]
	set above [lindex $labSpec 2]
	.panel create text [expr {$width-30}] [expr {$height-$above}] \
		-fill $color  -text $label -tag $label
	.panel create rect [expr {$width-60}] [expr {$height-$above-10}] \
		$width [expr {$height-$above+10}] \
		-fill "" -outline "" -tag $label
    }

    set lightX [expr {$width/2+30}]
    set lightY [expr {$height-83-30}]
    .panel create bitmap  $lightX $lightY \
	    -bitmap @$bitmaps/light.xbm -foreground MidnightBlue

    foreach descr {
	{red    firebrick -70}
	{yellow goldenrod -38}
	{green  DarkGreen  -6}
    } {
	set onColor  [lindex $descr 0]
	set offColor [lindex $descr 1]
	set diffY    [lindex $descr 2]

	.panel create bitmap $lightX [expr {$lightY+$diffY}] \
		-bitmap @$bitmaps/shade.xbm -foreground blue
	.panel create bitmap $lightX [expr {$lightY+$diffY+13}] \
		-bitmap @$bitmaps/ring.xbm -foreground gray60 -tag ring:$onColor
	.panel create bitmap $lightX [expr {$lightY+$diffY+13}] \
		-bitmap @$bitmaps/bulb.xbm -foreground $offColor -tag bulb:$onColor
	.panel create bitmap $lightX [expr {$lightY+$diffY+13}] \
		-bitmap @$bitmaps/shine.xbm -foreground white -tag shine:$onColor
	.panel lower shine:$onColor
	.panel create oval [expr {$lightX-13}] [expr {$lightY+$diffY}]  \
		[expr {$lightX+13}] [expr {$lightY+$diffY+26}] \
		-fill "" -outline "" -tag $onColor

    }

    foreach comm {red yellow green Quit Info Help} {
	.panel bind $comm <Enter> {.panel config -cursor hand2}
	.panel bind $comm <Leave> {.panel config -cursor top_left_arrow}
    }
    .panel bind red    <1>   {changeStateTo offline; break}
    .panel bind yellow <1>   {changeStateTo cold;    break}
    .panel bind green  <1>   {changeStateTo warm;    break}
    .panel bind Quit   <1>   {quitPanel 1;           break}
    .panel bind Info   <1>   {infoWindow;            break}
    .panel bind Help   <1>   {callHtmlHelp;          break}
    bind  .panel <Control-c> {quitPanel 1;           break}
    bind  .panel <F1>        {callHtmlHelp;          break}
    bind  .panel <F3>        {quitPanel 1;           break}
    focus .panel

    .panel lower light
    .panel lower shine

    set useNetscape [string match {[Tt]*} \
	    [option get .panel netscapeHelp NetscapeHelp]]

    update
}

# ---------------------------------------------------------------------
# trySetFont: Try to set the first font of a given list, that is
#             available on the current X server, without annoy the#
#             user with any error messages.
# ---------------------------------------------------------------------
proc trySetFont {c tag args} {
    foreach font $args {
	if {![catch {$c itemconfigure $tag -font $font}]} return
    }
}

# ---------------------------------------------------------------------
# panelShowMsg: Displays the given message on the picture. Without
#               parameters it removes any message from the canvas.
# ---------------------------------------------------------------------
proc panelShowMsg {{msg ""}} {
    .panel itemconfigure msg -text $msg
    update
}

# ---------------------------------------------------------------------
# panelShowDBState: The procedure to switch the traffic light, so that
#                   it corresponds to the state of the databas server.
#                   It switches the correct light on, and activates the
#                   corresponding (dimmed) lights.
#                   It also displays the name and state of the database.
# ---------------------------------------------------------------------
proc panelShowDBState {} {
    global connect currVal

    foreach colors {{red firebrick} {yellow goldenrod} {green DarkGreen}} {
	set onColor  [lindex $colors 0]
	set offColor [lindex $colors 1]
	.panel itemconfigure ring:$onColor -foreground gray60
	.panel itemconfigure bulb:$onColor -foreground $offColor
	.panel lower shine:$onColor
    }
    foreach entry {Offline Cold Warm} {
	confMenueEntry .m $entry -state normal
    }

    set currentState [dbState]

    switch $currentState {
	offline {set color red;    set entry  Offline}
	cold    {set color yellow; set entry Cold}
	warm    {set color green;  set entry Warm}
	default {
	    .panel itemconfigure dbState -text "$currentState"
	    error $currentState
	}
    }

    .panel itemconfigure dbState -text \
	    "Database $connect(serverdb) is $currentState"

    .panel itemconfigure ring:$color -foreground white
    .panel itemconfigure bulb:$color -foreground $color
    .panel raise shine:$color
    confMenueEntry .m $entry -state disabled
    
    set currVal(state) [string toupper $currentState]
    return $currentState
}

# ---------------------------------------------------------------------
# changeToState: The callback procedure for all colors of the panel
#                traffic light. The actual color will be given coded
#                as (warm|cold|offline) in parameter targetState.
# ---------------------------------------------------------------------
proc changeStateTo {targetState} {
    global connect

    if {[catch panelShowDBState currentState]} return

    if {![string compare $currentState $targetState]} return

    switch $currentState {
	offline {
	    if {[catch {externalProg "x_start $connect(serverdb)" \
		    		     "Starting '$connect(serverdb)'" cold} \
			currentState]} return
	}
	warm {
	    if {![string compare [autosaveShow] ENABLED]} {
		catch {utilityCmd "AUTOSAVE END" "Disabling autosave"}
	    }
	    if {[catch {utilityCmd shutdown \
				   "Shutdown '$connect(serverdb)'"} \
			currentState]} return
	}
    }

    if {![string compare $currentState $targetState]} return

    switch $targetState {
	warm {
	    catch {utilityCmd restart "Restarting '$connect(serverdb)'"}
	}
	offline {
	    catch {externalProg "x_stop $connect(serverdb)" \
				"Stopping '$connect(serverdb)'" offline}
	}
    }
}

# ---------------------------------------------------------------------
# externalProg: Execute the given external program while displaying an
#               info message.
#               Return the resulting database state (offline|cold).
# ---------------------------------------------------------------------
proc externalProg {prog msg targetState} {

    panelShowMsg "$msg,\nplease wait"
    after 100 "catch {exec $prog}; set progReady 1"
    tkwait variable progReady

    if {[catch panelShowDBState currentState]} {
	return
    }
    if {[string compare $currentState $targetState]} {
	panelShowMsg "Something went wrong,\nsee diag file"
	error ""
    }
    panelShowMsg
    return $currentState
}

# ---------------------------------------------------------------------
# utilityCmd: Execute the given utility command while displaying an
#             info message.
#             Return the resulting database state (cold|warm).
# ---------------------------------------------------------------------
proc utilityCmd {cmd msg} {
    panelShowMsg "$msg,\nplease wait"
    set error [sendKernel $cmd]
    if {$error} {
	panelShowMsg "Error $error while $cmd"
	return ""
    }
    panelShowMsg
    if {[catch panelShowDBState currentState]} {return ""}
    return $currentState
}

# ---------------------------------------------------------------------
# quitPanel: quits this program, not after asking, whether the
#            database should be stopped, too...
# ---------------------------------------------------------------------
proc quitPanel {askShutdown} {
    global connect

    if {$askShutdown} {
	switch [dbState] cold - warm {
	    if {[tk_dialog .quit "Stop" \
		    "Should database '$connect(serverdb)' be stopped, too?" \
		    questhead 0 "Only quit" "Stop database, then Quit"]} {
		update
		changeStateTo offline
	    }
	}
    }
    exit 0
}

# ---------------------------------------------------------------------
# createPopup: Create a popup menu.
# ---------------------------------------------------------------------
proc createPopup {} {
    global currVal

    menu .m

    .m add cascade -label Help -menu .m.help
    menu .m.help
    .m.help add command -label "Help..."  -command callHtmlHelp -und 0 -acc F1
    .m.help add command -label "About..." -command aboutHelp    -und 0
    .m.help add check   -label "Use Netscape" -variable useNetscape -und 0
    .m add separator

    .m add command -label "Change State of DB" -state disabled
    foreach state {Offline Cold Warm} {
	.m add command -label $state -und 0 \
		-command "changeStateTo [string tolower $state]"
    }

    .m add separator
    .m add command -label "Info Sheet" -und 0 -command "infoWindow"
    .m add cascade -label "Refresh"    -menu .m.refr

    .m add separator
    .m add command -label "Quit Panel" -und 0 -command "quitPanel 0"

    .panel bind bg <1> "tk_popup .m %X %Y"

    menu .m.refr
    .m.refr add radio -label "10 sec" -variable currVal(refresh) -value  10000
    .m.refr add radio -label "30 sec" -variable currVal(refresh) -value  30000
    .m.refr add radio -label " 2 min" -variable currVal(refresh) -value 120000
    .m.refr add radio -label "10 min" -variable currVal(refresh) -value 600000
    set currVal(refresh) 10000
}

# ---------------------------------------------------------------------
# aboutHelp: The notorius About... dialog.
# ---------------------------------------------------------------------
proc aboutHelp {} {
    tk_dialog .about "About Panel" "Adabas D: Panel
Version 0.5\n\nOctober 1996\n\nChristian Krone
krischan@cs.tu-berlin.de" "" 0 Ok
}

# ---------------------------------------------------------------------
# infoWindow: Creates a toplevel window of its own with a control like
#             outfit. Since most of this info sheet is in infoSheet.tcl,
#             here we are creating only the frame, and call a procedure
#             to fill the values.
# ---------------------------------------------------------------------
proc infoWindow {} {
    global connect

    if {[winfo exists .info]} {
	if {![string compare [wm state .info] iconic]} {wm deiconify .info}
	raise .info
	return
    }

    toplevel .info
    wm title    .info "Adabas D: Info Sheet for $connect(serverdb)"
    wm iconname .info "$connect(serverdb) Info Sheet"

    button .info.b -text Dismiss -command "dismissInfo"
    pack   .info.b -side bottom -expand 1 -fill x

    
    if {[catch {mkInfoSheet .info.i} msg]} {
	destroy .info
	tk_dialog .info "No Info Sheet" \
		"No info sheet due to the following error:\n$msg" \
		questhead 0 Sorry
	return
    }

    setCurrVal
}

# ---------------------------------------------------------------------
# setCurrVal: A procedure to fill the values of the Info Sheet. First
#             it determines the state of the database server
#             (offline,cold,warm). If the database is warm, it sends
#             additional selects, to adjust the labels and bars.
#             At last it reschedules itself to be called again, when
#             the refresh timespan is elapsed (by means of 'after').
#             Note the SELECT INTOS(!) at the core of this procedure.
# ---------------------------------------------------------------------
proc setCurrVal {} {
    global currVal connect adamsg

    set currVal(state) WARM
    if {![controlConnect]} {
        if {[info exists adamsg(errortxt)]} {
	    if {$adamsg(rc) == -8000} {
		set currVal(state) COLD
	    } else {
		if {$adamsg(rc) == -4008} {
		    set connect(error) "No valid Control user given"
		} else {
		    set connect(error) "$adamsg(rc): $adamsg(errortxt)"
		}
		getDbnameUserPassword "Adabas D Panel Connect" controlConnect
		if {![info exists connect(db)]} {
		    error "Invalid Control User"
		    return
		}
	    }
	} else {
	    set currVal(state) OFFLINE
	    return
        }
    }

    if {![string compare $currVal(state) WARM]} {
	set l $connect(db)
	set c $connect(cursor)

	regexp {\(for Adabas D (.*)\)} $adamsg(version) match version
	set vers62 [expr {$version >= 6.2}]

	# Switch on monitoring, if not already turned on.
	if {$vers62} {
	    adasql $c {SELECT value INTO :monitorOnOff
                         FROM sysdd.server_db_state
                         WHERE description = 'MONITORING'}
        } else {
	    adasql $c {SELECT value INTO :monitorOnOff
                         FROM sysdd.serverdbstate
                         WHERE description = 'MONITORING'}
        }
	if {[string compare $monitorOnOff ON]} {
	    adasql $c {MONITOR ON}
	}

	if {$vers62} {
	    adasql $c {SELECT used_perm_pages, pct_perm_used,
                              used_log_pages,  pct_logused,
                              completed_log_segm
                         INTO :currVal(data), :currVal(data_pct),
                              :currVal(log),  :currVal(log_pct),
                              :currVal(logSegments)
                         FROM sysdd.serverdb_stats}
        } else {
	    adasql $c {SELECT used_perm_pages, pct_used_perm,
                              used_log_pages,  pct_used_log,
                              completed_log_segm
                         INTO :currVal(data), :currVal(data_pct),
                              :currVal(log),  :currVal(log_pct),
                              :currVal(logSegments)
                         FROM sysdd.serverdbstatistics}
        }
	if {![string compare $currVal(logSegments) 0]} {
	    set currVal(logSegments) "NOT COMPLETED"
	} else {
	    set currVal(logSegments) "COMPLETED"
	}

	adasql $c {SELECT value INTO :maxUser FROM sysdd.xparameters
	  	     WHERE description = 'MAXUSERTASKS'}

        adasql $c {SELECT COUNT (session) INTO :currVal(sessions)
                     FROM sysdd.transactions}
        set currVal(sessions_pct) [expr {$currVal(sessions)*100/$maxUser}]

        adasql $c {SELECT value INTO :currVal(reads) FROM sysdd.monitor
                     WHERE type = 'PAGES'
                       AND description = 'VIRTUAL  READS'}

        adasql $c {SELECT value INTO :currVal(writes) FROM sysdd.monitor
                     WHERE type = 'PAGES'
                       AND description = 'VIRTUAL  WRITES'}

        adasql $c {SELECT value INTO :currVal(sqlCommands) FROM sysdd.monitor
                     WHERE type = 'TRANS'
                       AND description = 'SQL COMMANDS'}

        adasql $c {SELECT value INTO :currVal(hitrate) FROM sysdd.monitor
                     WHERE type = 'CACHES'
                       AND description = 'DATA CACHE HIT RATE (%)'}

        adasql $c {SELECT value INTO :currVal(converter) FROM sysdd.monitor
                     WHERE type = 'CACHES'
                       AND description = 'CONVERTER CACHE HIT RATE (%)'}

        set currVal(stateAt) [clock format [clock seconds] \
		-format "%Y-%m-%d %H:%M:%S"]
		  
        adaclose  $c
        adalogoff $l

	set currVal(autosaveLog) [autosaveShow]
    }
    set currVal(afterID) [after $currVal(refresh) setCurrVal]
}

# ---------------------------------------------------------------------
# dismissInfo: Removes the Info Sheet and cancels any outstanding refresh.
# ---------------------------------------------------------------------
proc dismissInfo {} {
    global currVal

    catch {after cancel $currVal(afterID)}
    destroy .info
}

proc autosaveShow {} {
    global connect

    set ret DISABLED
    if {[utilConnect]} {
	if {![catch {adautil $connect(db) "AUTOSAVE SHOW"}]} {
	    set ret ENABLED
	}
	adalogoff $connect(db)
    }
    return $ret
}

# ---------------------------------------------------------------------
# dbState: The procedure, that communicates with the database to
#          evaluate its state and return it coded as (warm|cold|offline).
#          It expects some valuable information in dbname and user/passwd.
# ---------------------------------------------------------------------
proc dbState {} {
    global connect adamsg argv

    if {[catch "adalogon $connect(param) -service utility \
	    				 -serverdb $connect(serverdb)" log]} {
        if {[info exists adamsg(errortxt)]} {
	    if {$adamsg(rc) == -4008} {
		set connect(error) "No valid Control user given"
	    } else {
		set connect(error) "$adamsg(rc): $adamsg(errortxt)"
	    }
	    getDbnameUserPassword "Adabas D Panel Connect" utilConnect
	    if {[info exists connect(db)]} {
	        set log $connect(db)
	    } else {
	        error "Invalid Control User"
	    }
	} else {
	    return offline
	}
    }
    set stateOut [adautil $log "STATE"]
    catch {adautil $log "COMMIT WORK RELEASE"}
    adalogoff $log
    return [lindex $stateOut 0]
}

# ---------------------------------------------------------------------
# sendKernel: send the given utility command to the database kernel.
# ---------------------------------------------------------------------
proc sendKernel {command} {
    global connect

    if {[catch "adalogon $connect(param) -service utility \
	    				 -serverdb $connect(serverdb)" db]} {
	return "permission denied"
    }

    catch {adautil $db $command} commandOut
    catch {adalogoff $db}
    return 0
}

# ---------------------------------------------------------------------
# utilConnect: Try to connect with the given user key or user,password
#              combination as utility service.
# ---------------------------------------------------------------------
proc utilConnect {} {
    global connect adamsg

    if {[info exists connect(userkey)] && ![info exists connect(clearpw)]} {
        set connect(param) ",$connect(userkey)"
    } else {
	set connect(param) "$connect(user),$connect(clearpw)"
    }

    if {[catch "adalogon $connect(param) -service utility \
	    				 -serverdb $connect(serverdb)" db]} {
	if {[info exists adamsg(rc)]} {
	    if {$adamsg(rc) == -4008} {
		set connect(error) "No valid Control user given"
	    } else {
		set connect(error) "$adamsg(rc): $db"
	    }
	} else {
	    set connect(error) $db
	}
	return 0
    }

    set connect(db) $db
    return 1
}

# ---------------------------------------------------------------------
# controlConnect: Try to connect with the given user key or user,password
#                 combination as control user.
# ---------------------------------------------------------------------
proc controlConnect {} {
    global connect adamsg

    if {[info exists connect(userkey)] && ![info exists connect(clearpw)]} {
        set connect(param) ",$connect(userkey)"
    } else {
	set connect(param) "$connect(user),$connect(clearpw)"
    }
    catch {unset connect(servernode)}

    if {[catch "adalogon $connect(param) -service control \
	    				 -serverdb $connect(serverdb)" db]} {
	if {[info exists adamsg(rc)]} {
	    if {$adamsg(rc) == -4008} {
		set connect(error) "No valid Control user given"
	    } else {
		set connect(error) "$adamsg(rc): $db"
	    }
	} else {
	    set connect(error) $db
	}
	return 0
    }

    set c [adaopen $db]
    if {[catch {adasql $c {SELECT value FROM sysdd.xparameters
                             WHERE description='CONTROLUSERID'
			       AND value=USER}}]} {
        adaclose  $c
	adalogoff $db
	set adamsg(rc)       -4008
	set adamsg(errortxt) "UNKNOWN USER/PASSWORD COMBINATION"
	set connect(error)   "No valid Control user given"
	return 0
    }

    set connect(db)     $db
    set connect(cursor) $c
    return 1
}

# ---------------------------------------------------------------------
# checkLocal: Try to determine, if the given connect data really stays
#             on this host. This procedure is very dumb(!)
# ---------------------------------------------------------------------
proc checkLocal {} {
    global connect

    # But make shure, that at least a user key or user name was given...
    if {(![info exists connect(userkey)]
         || ![string length $connect(userkey)])
        && ![string length $connect(user)]} {
	set connect(error) "Enter user name and password"
	return 0
    }

    if {[string length $connect(servernode)]} {
	set connect(error) "No remote panel possible"
	return 0
    }

    if {[info exists connect(userkey)] && ![info exists connect(clearpw)]} {
        set connect(param) ",$connect(userkey)"
    } else {
	set connect(param) "$connect(user),$connect(clearpw)"
    }

    return 1
}

# ---------------------------------------------------------------------
# confMenueEntry: Since the menuentries in tearoffs are not updated
#                 together with their original, you should use this
#                 procedure instead of entryconfigure directly.
# ---------------------------------------------------------------------
proc confMenueEntry {m index args} {
    set tearOffs [info commands .tearoff*]
    if {[llength $tearOffs]} {
        set m "$m $tearOffs"
    }
    foreach menu $m {
        catch {eval $menu entryconfigure [list $index] $args}
    }
}

# ---------------------------------------------------------------------
# callHtmlHelp: Small wrapper around htmlHelp.
# ---------------------------------------------------------------------
proc callHtmlHelp {} {
    global adabastcl_library useNetscape

    htmlHelp $adabastcl_library/manuals/panel.html $useNetscape
}

# ---------------------------------------------------------------------
# Mainprogram
# ---------------------------------------------------------------------
wm withdraw .
loadAdabastcl
createPanel
createPopup
getDbnameUserPassword "Adabas D Panel Connect" checkLocal
wm deiconify .
update
if {![catch panelShowDBState currentState]} {
    after 100 {changeStateTo warm}
}

# ---------------------------------------------------------------------
# Some stuff for editing with emacs...
# ---------------------------------------------------------------------
# Local Variables: 
# mode: tcl
# End: 
